import subprocess
import platform


def getBssid():
  if platform.system() == "Windows":
    # Run the netsh command to get the BSSID of the connected WiFi network
    output = subprocess.check_output(["netsh", "wlan", "show", "interfaces"])
    # Extract the BSSID from the output
    return output.decode("gbk").split("BSSID")[1].split()[1]
  elif platform.system() == "Linux":
    # Run the iwconfig command to get the BSSID of the connected WiFi network
    output = subprocess.check_output(["iwconfig"])
    # Extract the BSSID from the output
    return output.decode().split("Access Point: ")[1].split()[0]


def getIPAndMac():
  import netifaces

  # Get the IP address and MAC address of the default network interface
  default_interface = netifaces.gateways()['default'][netifaces.AF_INET][1]
  addresses = netifaces.ifaddresses(default_interface)[netifaces.AF_INET]

  # Find the first non-loopback IP address
  ip_address = next((addr['addr'] for addr in addresses if not addr['addr'].startswith('127.')), None)

  # Get the MAC address
  mac_address = netifaces.ifaddresses(default_interface)[netifaces.AF_LINK][0]['addr']

  return ip_address, mac_address


def getNetInfo():
  ip, mac = getIPAndMac()
  bssid = getBssid()
  return ip, mac, bssid
    



from paho.mqtt.client import Client as MQTTClient
from ..core.scheduleService import ScheduleBase
from .. import config

# from ..router import httpRouter

import json
import _thread
from ..utils import logutil
import uuid

log = logutil.get_logger(__name__)

# 泛在云mqtt服务
class MqttService(ScheduleBase):
  #mqtt客户端
  client = None
  deviceId = "Unknown Device"
  lock=_thread.allocate_lock()

  
  """
  * deviceId   泛在平台设备id
  * address   mqtt地址
  * port  mqtt端口号
  """
  def __init__(self,deviceId:str,address:str=config.MQTT_ADDRESS ,port=config.MQTT_PORT):
    print(address)
    print(port)
    self.deviceId = deviceId
    if address == config.MQTT_ADDRESS and port == config.MQTT_PORT:
        self._SystemMqttInit()
    else:
        self._MqttInit(address,port)
    print("MQTT Init Success")

    

  '''
  * 消息监听
  '''
  def check_msg(self):
    try:
      self.client.loop()
    except Exception as e:
      log.error(f"消息监听失败: {str(e)}")
    

  """
  * mqtt初始化
  """
  def _SystemMqttInit(self):
    log.info("init mqtt")
    self.client = MQTTClient(self.deviceId,clean_session=True)  # 创建MQTT对象
    willMsg = {
      "deviceInstanceId":config.DEVICE_ID,
      "serviceUrl": "systemService",
	    "versionCode": 0,
	    "eventId": "OFFLINE",
	    "msg": None
      }
    self.client.will_set(topic = config.ON_EVENT_TOPIC, payload=json.dumps(willMsg, ensure_ascii=False))

    # 设置连接和消息回调函数
    self.client.on_connect = self.on_connect
    self.client.on_message = self.sub_cb

    self.client.connect(host = config.MQTT_ADDRESS, port = config.MQTT_PORT, keepalive = 60)
    # 开始循环处理网络流量和消息回调
    self.client.loop()


  """
  * mqtt初始化
  """
  def _MqttInit(self,address:str,port:int):
    self.client = MQTTClient(self.deviceId + str(uuid.uuid1()))  # 创建MQTT对象
    self.client.connect(host = address, port = port, keepalive = 60)


  """
  * 发送消息
  * topIc 主题消息 字符串消息
  * message 一般为一个json格式的字符串
  """
  def sendMsg(self,topic:str,message:str):
    if self.lock.acquire():
        try:
            self.client.publish(topic,message,qos=2)
        except :
            self._SystemMqttInit()
        finally :
            self.lock.release()

  def ping(self):
    self.client.ping()


  """
  * 消息回调
  """
  def sub_cb(self, client, userdata, msg):
    try:
      from ..router import mqttRouter
      topic = msg.topic
      msg = msg.payload.decode()
      mqttRouter.topicRouter(topic,msg)
    except Exception as e:
      log.error(f" {str(e)}")
  
  def close(self):
    self.client.disconnect()


  # 定义 MQTT 客户端的回调函数
  def on_connect(self, client, userdata, flags, rc):
    if rc == 0:
      # 消息订阅    
      client.subscribe(config.INVOCATION_SERVICE_EXECUTE_TOPIC)
      client.subscribe(config.UPDATE_SERVICE_TOPIC)
      log.info("系统mqtt初始化成功")
    else:
      log.error("Failed to connect, return code {0}".format(rc))







import sys
import os
from .utils import logutil
from .utils import netUtils
log = logutil.get_logger(__name__)
# 获取主程序所在的目录
program_dir = os.path.dirname(sys.argv[0])
program_path = os.path.abspath(program_dir)
log.info("程序所在目录:"+ program_path)
# 获取当前文件的绝对路径
model_file_path = os.path.abspath(__file__)
# 获取当前文件所在的目录
model_directory = os.path.dirname(model_file_path)
log.info("模块所在目录:"+ model_directory)

from typing import Callable
from .systemService.mqttService import MqttService
from .core import applicationContext
from .core.loader.proxyLoader import ProxyLoader
from .core.loader.serversLoader import ServiceLoader
from .utils import pathUtils as PathUtil
from .core.scheduleService import ScheduleService
from .core.job.mqttCheckMessageJob import MqttCheckMessageJob
from .core.job.contextUpLoadJob import ContextUploadJob
from .core.job.eventListenerJob import EventListenerJob
from . import config
from .utils import netUtils
from .core.loader import thingProxyFuncLoader
import  json
from .core.serverInvoker import EventInvoker
from .core.loader.hardwareInfoLoader import HardwareInfoLoader
from .core.sender.hardwareUpLoadSender import HardwareUpLoadSender


class ProxySDK:
  # 泛在web服务
  webServer = None

  # 泛在mqtt服务
  mqttClient = None


  """
  * 全部功能初始化
  * deviceId   泛在平台设备id
  * setThingModel 物模型读写函数
  * wlan 网络对象 micropython连接网络和存储网络信息的对象
  * deviceTypeId 所属设备类型的id
  * hardwareInfoLoader 硬件信息加载对象
  * proxyPath 代理所在路径（用户自定义代理）（默认proxy文件夹）
  * servicePath 服务存储到哪个位置（默认"/service"）
  * BASE_PATH 程序基本路径（默认为根目录）
  * SDK_PATH SDK所在路径（默认upip下载路径）
  * openHttp 是否开启局域网调用功能（默认开启）
  """
  def __init__(self,deviceId:str,setThingModel:Callable,
                deviceTypeId:str,hardwareInfoLoader:HardwareInfoLoader,proxyPath = "proxy",servicePath = "service",
                BASE_PATH=program_path,SDK_PATH=model_directory,openHttp=True):
    proxyPath = PathUtil.joinPath(BASE_PATH,proxyPath)
    # 初始化配置文件
    config.initTopic(deviceId)
    config.initPath(BASE_PATH,servicePath,proxyPath)
    # 配置物模型读写函数
    thingProxyFuncLoader.setFunc(setThingModel)
    # 读取proxy字典
    #   加载用户代理
    applicationContext.proxyDict = ProxyLoader.loadProxyDict(proxyPath,BASE_PATH)
    #   加载系统代理
    applicationContext.proxyDict.update(ProxyLoader.loadProxyDict( PathUtil.joinPath(SDK_PATH,"proxy"),BASE_PATH))
    # 解析并存储服务字典
    applicationContext.serviceDict = ServiceLoader.LoadServiceDict(servicePath,BASE_PATH)
    # 保存硬件信息加载回调
    applicationContext.hardwareInfoLoader = hardwareInfoLoader
    # 初始化mqtt服务
    applicationContext.pervasiveMqttClient = MqttService(deviceId)


    # mqtt上报设备状态信息
    serviceList = applicationContext.getServiceListInfo()
    ip, mac, bssid = netUtils.getNetInfo()
    deviceInfo = {}
    deviceInfo["macAddress"] = mac.lower()
    deviceInfo["routerMacAddress"] = bssid.lower()
    deviceInfo["ipAddress"] = ip.lower()
    deviceInfo["serviceList"] = serviceList
    deviceInfo["deviceTypeId"] = deviceTypeId
    log.info(deviceInfo)
    applicationContext.stateInfo = deviceInfo
    config.deviceInfo = deviceInfo

    applicationContext.pervasiveMqttClient.sendMsg(config.UPLOAD_SERVICE_LIST_TOPIC,json.dumps(deviceInfo, ensure_ascii=False))

    # 初始化定时任务  
    tasks = []
    # tasks.append(MqttHealthJob(intervalTime=60000)) 
    tasks.append(ContextUploadJob(intervalTime=10000))
    tasks.append(MqttCheckMessageJob(intervalTime=500))
    tasks.append(EventListenerJob(intervalTime=5000))
    applicationContext.taskServer = ScheduleService(tasks)
    # 初始化服web服务 
    if(openHttp):
      from .systemService import httpService
      self.webServer = httpService.webServerInit()
    EventInvoker.invokeSystemEvent("ONLINE")
    HardwareUpLoadSender.send()

  
    



    

    
  


# from machine import Timer

# 最小间隔时间
MIN_INTERVAL_TIME = 50
import sched
import threading
import time
from ..utils import logutil

log = logutil.get_logger(__name__)
'''
任务基类，所有定时任务都需要继承此类
==注意== 此类所有时间单位均为ms ; 1s = 1000ms
'''
class ScheduleBase:
  #下次执行时间
  nextRunTime = 0
  #间隔时间
  intervalTime = 1000

  log = None

  def __init__(self,intervalTime=1000) -> None:
    self.intervalTime =intervalTime<50 and 50 or intervalTime
    self.log = logutil.get_logger(__name__)

  def setIntervalTime(self,intervalTime):
    self.intervalTime=intervalTime

  def tryHandleInterrupt(self):
    try:
      self.handleInterrupt()
    except Exception as e:
      log.error(str(e))


  def handleInterrupt(self):
    print("please rewrite the handleInterrupt")

'''
定时任务类
'''
class ScheduleService:
  tasks = []
  divisor = 1000
  scheduler = None
  task = None


  def __init__(self,tasks):
    self.tasks = tasks

    # 创建调度器
    self.scheduler = sched.scheduler(time.time, time.sleep)
    task = threading.Thread(target=self.task_thread)
    task.start()

    # #intervalTime使用最大公约数来决定轮巡间隔
    # intervalTime = len(tasks)>0 and tasks[0].intervalTime or 1
    # for task in tasks:
    #   intervalTime = gcd(task.intervalTime,intervalTime)
    # # 限幅  如果执行的间隔时间小于50毫秒则等于50毫秒
    # self.intervalTime =intervalTime<50 and 50 or intervalTime
    # print("ScheduleService Init Success intervalTime")


  # def handleInterrupt(self,timer):
  #   for task in self.tasks:
  #     task.nextRunTime = task.nextRunTime - self.intervalTime
  #     if(task.nextRunTime<=0):
  #       task.handleInterrupt()
  #       task.nextRunTime = task.intervalTime

  def handleInterrupt(self,schedule:ScheduleBase):
    schedule.tryHandleInterrupt()

    self.scheduler.enter(schedule.intervalTime/self.divisor, 1, self.handleInterrupt, (schedule,))

  
  # 创建并启动线程
  def task_thread(self):
      # 开始执行任务  
      for task in self.tasks:
        self.scheduler.enter(3, 1, self.handleInterrupt, (task,))
      self.scheduler.run()







from ..pojo.serviceBean import ServiceBean
from . import systemEventFactory
from .. import config

'''
* 系统自带的service
* 系统服务工厂
'''
class SystemServiceFactory:
  def loadSystemService()-> ServiceBean:
    res = ServiceBean()
    res.url = "systemService"
    res.name = "systemService"
    res.versionCode = 0
    eventBeanDict = {}
    
    online = systemEventFactory.createSystemEvent("ONLINE",config.ON_EVENT_TOPIC)
    offline = systemEventFactory.createSystemEvent("OFFLINE",config.ON_EVENT_TOPIC)

    eventBeanDict[online.eventId] = online
    eventBeanDict[offline.eventId] = offline

    res.eventBeanDict = eventBeanDict
    return res


import os
import sys
from ...utils import logutil as logUtil
import importlib
from ...utils import pathUtils as PathUtil
# import importlib.util

log = logUtil.get_logger(__name__)
'''
代理加载工具
''' 
class ProxyLoader:
  """
  * 读取模块字典
  * modelPath   model所在路径,相对于src目录
  """
  def loadProxyDict(proxyDirPath,BASE_PATH): 
    proxyDict = {}
    absolutePath=proxyDirPath
    proxyList = []
    try:
      proxyList = os.listdir(absolutePath)
    except Exception as e:
      log.warning("代理文件夹不存在"+absolutePath)
      return {}
    for proxyName in proxyList:
      try:
        if proxyName.find(".py")>0:
          module_path = PathUtil.joinPath(proxyDirPath, proxyName)
          log.debug(module_path)
          proxyDict = ProxyLoader.load_proxy_dict(module_path,proxyDict)
      except Exception as e:
        log.error(proxyDirPath+"  "+proxyName+"代理加载失败"+str(e))
    print("Load ModelDict Success")
    return proxyDict


  def load_proxy_dict(file_path,proxyDict:dict):
    try:
        file_name = os.path.basename(file_path)
        file_name = os.path.splitext(file_name)[0]
        spec = importlib.util.spec_from_file_location(file_name, file_path)

        module = importlib.util.module_from_spec(spec)

        spec.loader.exec_module(module)

        proxyDict[file_name] = module

    except Exception as e:
      log.error(f"加载代理 {file_path} 失败: {str(e)}")

    return proxyDict







import requests
import os
import shutil
from ..utils import logutil

log = logutil.get_logger(__name__)

'''
 删除路径及其中文件
'''
def removePath(path:str):
    shutil.rmtree(path)
    
    
'''
 下载文件
 url:文件链接
 path:目标路径（包含文件名）
'''
def downloadFile(url:str,path:str):
    # 确认路径存在    
    fileFolderPath = os.path.dirname(path)
    
    # 删除已存在的文件     
    try:
        removePath(path)
    except:
        pass
    create_path(fileFolderPath)
    # 下载
    res = requests.get(url)
    with open(path,"wb") as code:
        code.write(res.content)

'''
  创建路径
'''
def create_path(path):
    try:
        os.makedirs(path, exist_ok=True)
        log.info(f"Path '{path}' created.")
    except OSError as error:
        log.error(f"creating path '{path}': {error}")



readFunc = None
writerFunc = None

def setFunc(thingModelClass):
    global readFunc
    global writerFunc
    readFunc = thingModelClass.readThingModel
    writerFunc = thingModelClass.setThingModel



from ..core.serverInvoker import CapabilityInvoker
from .. import config
import json
from ..core import applicationContext
from ..utils import TimeUtil
from ..utils import logutil
from ..systemService import updateService
import sys

routerLog = logutil.get_logger(__name__+" MQTT router")
serviceLog = logutil.get_logger(__name__ + " service invoke")
updateLog = logutil.get_logger(__name__ + " update invoke")

def topicRouter(topic,msg):
  
  topic = topic.split("/")
  print(topic)

  # 调用能力接口
  if topic[2] == config.INCOCATION_ENUM:
    msgJson = json.loads(msg)
    msgId = msgJson.get("messageId","null")
    runSuccess = False
    responce = None
    routerLog.info("======mqtt server invoke======",msgId)
    try:
      responce = CapabilityInvoker.invokeCapability(topic[-2],topic[-1],msgJson["param"])
      runSuccess = True
    except Exception as e:
      serviceLog.error(str(e))
      runSuccess = False
      responce = {"errorMsg":str(e)}
    routerLog.info("====callback====")
    # 拼接topic
    callbackTopic = config.INVOCATION_SERVICE_CALLBACK_TOPIC
    callbackTopic = callbackTopic.replace("${service}",topic[-2])
    callbackTopic = callbackTopic.replace("${capability}",topic[-1])
    callbackTopic = callbackTopic.replace("${msgid}",msgId)
    if (msgId==None):
        serviceLog.error("The message body does not contain the 'messageId'")
    routerLog.info(callbackTopic)
    # 拼接msg
    msgDict = {
      "success" : runSuccess,
      "data" : responce,
      "timestamp": TimeUtil.getTimerMs()
    }
    routerLog.info(msgDict)
    applicationContext.pervasiveMqttClient.sendMsg(callbackTopic,json.dumps(msgDict, ensure_ascii=False))
  # 脚本更新能力
  elif topic[2] == config.UPDATE_ENUM:
    msgJson = json.loads(msg)
    print(msgJson)
    msgId = msgJson.get("messageId","null")
    isInstall = msgJson.get("installStatus",True)
    # 拼接topic
    callbackTopic = config.UPDATE_SERVICE_CALLBACK_TOPIC
    callbackTopic = callbackTopic.replace("${service}","download")
    callbackTopic = callbackTopic.replace("${msgid}",msgId)
    runSuccess = True
    responce = {}
    try :
      if isInstall:
        routerLog.info("======mqtt update service======")
        callbackTopic = callbackTopic.replace("${capability}","install")
        updateService.updateService(topic[4])
      else:
        routerLog.info("======mqtt delete service======")
        callbackTopic = callbackTopic.replace("${capability}","uninstall")
        updateService.deleteService(topic[4])
    except Exception as e:
        runSuccess = False
        updateLog.error(str(e))
        # sys.print_exception(e)
        responce = {"errorMsg":str(e)}
    msgDict = {
      "success" : runSuccess,
      "timestamp": TimeUtil.getTimerMs(),
      "data":responce
    }
    applicationContext.pervasiveMqttClient.sendMsg(callbackTopic,json.dumps(msgDict, ensure_ascii=False))
  else:
    routerLog.error("no topic path")
   





import logging
from .. import config



def get_logger(name: str):
    # 创建日志记录器
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # 创建控制台处理器
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)

    # 创建日志格式器
    formatter = logging.Formatter('%(asctime)s - %(name)s -  %(filename)s:%(lineno)d - %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)

    # 将处理器添加到日志记录器
    logger.addHandler(console_handler)
    return logger
all = (get_logger)


from ..scheduleService import ScheduleBase
from .. import applicationContext
from ... import config
from ...utils import TimeUtil
from ...proxy.thingProxy import thingProxy
import json


class ContextUploadJob(ScheduleBase):

  """
  * 发送情境消息
  """
  def handleInterrupt(self):
    # wlan = network.WLAN(network.STA_IF)
    thingModel = thingProxy({"operation":"read"}).handle(None)
    services = applicationContext.getServiceListInfo()
    # 拼接msg
    msgDict = {
      "timestamp": TimeUtil.getTimerMs(),
      # "lanIp" : wlan.ifconfig()[0],
      "thingModel":thingModel,
      "serviceList":services
    }
    applicationContext.pervasiveMqttClient.sendMsg(config.UPLOAD_CONTEXT_TOPIC,json.dumps(msgDict, ensure_ascii=False))
    
  def upLoadServiceList(self):
    services = applicationContext.getServiceListInfo()



    
        
    
    
    
    
    
    


import copy
import json
import os

from . import applicationContext
from .. import config
from .exception.serviceException import ServiceNotExistException
from .exception.serviceException import CapabilityNotExistException
from .exception.serviceException import ProxyNotExistException
from .exception.serviceException import InvokeServiceException
from .exception.commonException import invalidParamsException
from .exception.serviceException import ServiceNotExistException
from .exception.serviceException import EventNotExistException
from ..model.ujinja.source import TemplateEngine
from ..proxy.thingProxy import thingProxy
from ..pojo.serviceBean import ServiceBean
from ..pojo.eventBean import EventBean
from ..utils import pathUtils as PathUtil
from ..model.ujinja.source import TemplateEngine
from ..systemService.mqttService import MqttService


class CapabilityInvoker:
  '''
  * 调用能力
  * serviceId  服务id
  * capabilityId 能力id
  * postJsonDict 消息的参数字典
  '''
  def invokeCapability(serviceId:str,capabilityId:str,requestJsonDict:dict={}):
    requestJsonDict["system"] = applicationContext.getSystemInfo()
    # 特判物模型读写服务    
    if (serviceId == "thingService"):
        return thingProxy({"operation":capabilityId}).handle(requestJsonDict)
    # 从服务字典读取服务信息
    if (serviceId==None):
        raise invalidParamsException("id",serviceId)
    if (capabilityId == None):
        raise invalidParamsException("capabilityId",capabilityId)
    service = applicationContext.serviceDict.get(serviceId,None)
    if (service == None):
        raise ServiceNotExistException(serviceId)
    capability = service.capabilityBeanDict.get(capabilityId,None)
    if (capability==None):
        raise CapabilityNotExistException(capabilityId)
    # jinja2 输入消息转换
    runData = None
    try:
      if(capability.inputMapper != None):
        if(capability.inputLanguage == "jinja2"):
          inputTemplateEngine = TemplateEngine()
          inputTemplateEngine.load_template(capability.inputMapper)
          runData = inputTemplateEngine.render_template(requestJsonDict)
          print(runData)
    except Exception as e:
      raise InvokeServiceException("服务定义错误,输入消息映射错误,请联系服务开发者")
        
    initData =  copy.deepcopy(capability.attributes)
    # 如果是python代理则需要加载脚本所在位置
    if(capability.proxyType == "pythonServiceProxy"):
      scriptPath = PathUtil.joinPath(service.filePath,"script")
      initData["script"] = PathUtil.joinPath(scriptPath,initData["script"])
    response = CapabilityInvoker.invokeProxy(capability.proxyType,initData,runData)
    #输出消息转换
    if(capability.outputMapper != None):
      if(capability.outputLanguage == "jinja2"):
        outputTemplateEngine = TemplateEngine()
        outputTemplateEngine.load_template(capability.outputMapper)
        try:
          if(isinstance(response, str)):
            response = json.loads(response)
          response["system"] = applicationContext.getSystemInfo()
          response = outputTemplateEngine.render_template(response)
          print(type(response))
          print(response)
          if(isinstance(response, str)):
            response = json.loads(response)
          print(type(response))
          print(response)
        except Exception as e:
          raise InvokeServiceException(json.dumps({
            "response":response,
            "systemInfo":"服务定义错误,输出消息映射错误,请联系服务开发者"+str(e)
            }, ensure_ascii=False))
          
        print(response)
    return response


  '''
  * 运行代理
  * proxyName 代理名称
  * initData  初始化数据
  * runData   运行时数据
  '''
  def invokeProxy(proxyName:str,initData:any,runData):
      proxyModel = applicationContext.proxyDict.get(proxyName,None)
      if(proxyModel==None):
          raise ProxyNotExistException(proxyName)
      if(hasattr(proxyModel,proxyName)):
        proxyClass = getattr(proxyModel,proxyName)
        proxy = proxyClass(initData)
        return proxy.handle(runData)




'''
* 事件调用类
'''
class EventInvoker:

  '''
  *调用系统事件
  '''
  @staticmethod
  def invokeSystemEvent(eventId:str,eventArgs:dict = {}):
    EventInvoker.invokeEvent("systemService",eventId,eventArgs)


  '''
  * 调用事件
  * serviceId 服务id
  * eventId 事件id
  * eventArgs 事件参数字典
  '''
  @staticmethod
  def invokeEvent(serviceId:str,eventId:str,eventArgs:dict = {}):
    eventArgs["system"] = applicationContext.getSystemInfo()
    
    service:ServiceBean = applicationContext.serviceDict.get(serviceId,None)
    if service == None:
      raise ServiceNotExistException(serviceId)
    
    event:EventBean = service.eventBeanDict.get(eventId,None)
    if event == None:
      raise EventNotExistException(eventId,serviceId)
    
    eventMsg = None
    if event.messageMapper!=None:
      if(event.messageLanguage == "jinja2" or event.messageLanguage == None):
          inputTemplateEngine = TemplateEngine()
          inputTemplateEngine.load_template(event.messageMapper)
          eventMsg = inputTemplateEngine.render_template(eventArgs)
          print("== event msg Mapper == ")
          print(eventMsg)

    # 默认是mqtt的方式上报topic
    if event.protocolType == "MQTT" or event.protocolType == None:
      EventInvoker._MQTTEventSender(service,event,eventMsg)


  '''
  * mqtt事件发送
  '''
  @staticmethod
  def _MQTTEventSender(service:ServiceBean,event:EventBean,eventMsg:str):
    host:str = event.attributes.get("host",None)
    isSystemMqtt = False
    mqttClient = None

    # 默认使用系统的mqtt客户端
    if host == "" or host == None:
      isSystemMqtt = True
      mqttClient = applicationContext.pervasiveMqttClient
    else:
      addressList:list = host.split(":")
      addressList[1] = int(addressList[1])
      if addressList[0] == config.MQTT_ADDRESS and addressList[1] == config.MQTT_PORT:
        isSystemMqtt = True
        mqttClient = applicationContext.pervasiveMqttClient
      else:
        mqttClient = MqttService(''.join([('0'+hex(ord(os.urandom(1)))[2:])[-2:] for x in range(8)]),addressList[0],addressList[1])

    topic = event.attributes.get("topic",config.ON_EVENT_TOPIC)
    msgJsonDict = {
      "deviceInstanceId":config.DEVICE_ID,
      "serviceUrl": service.url,
      "versionCode": service.versionCode,
      "eventId":event.eventId,
      "msg": eventMsg
    }

    mqttClient.sendMsg(topic,json.dumps(msgJsonDict, ensure_ascii=False))
    # 如果非系统的mqtt客户端则关闭
    if not isSystemMqtt:
      mqttClient.close()


  def getEventMsgStr(service:ServiceBean,event:EventBean,eventMsg:str):
    msgJsonDict = {
      "deviceInstanceId":config.DEVICE_ID,
      "serviceUrl": service.url,
      "versionCode": service.versionCode,
      "eventId":event.eventId,
      "msg": eventMsg
    }
    return json.dumps(msgJsonDict, ensure_ascii=False)
    


  
    



# 事件Bean

class EventBean:
  
  #事件id
  eventId = None
  
  # 参数
  attributes = {}
  
  # 推送类型
  protocolType = None 
  
  # 输入模板
  messageMapper = None
  
  # 输出模板的语言
  messageLanguage = None

  
  def __init__(self) -> None:
    self.capabilityBeanDict = {}
    
  def __str__(self):
    return f"EventBean(EventId={self.eventId}, protocolType={self.protocolType}, attributes={self.attributes}, messageMapper={self.messageMapper}, messageLanguage={self.messageLanguage})"




import re
class TemplateEngine:
  FOR = "for"
  ENDFOR = "endfor"
  IF = "if"
  ENDIF = "endif"
  delimiterList=["==", ">=", "<=","!=",">","<","or", "and", "!"]
  delimiters = "("+"|".join(delimiterList)+")"

  def __init__(self):
      pass

  # 定义一个加载模板的方法，从文件中读取模板内容，并存入字典
  def load_template(self, strTemplates):
      self.template = strTemplates;

  # 渲染模板的方法
  # 输入 变量字典
  # 返回 渲染后的字符串
  def render_template(self, data: dict)->str:
    result = self.template
    # 获取标签位置
    start_index = self.findStartTag(result)
    print(start_index)

    while start_index != -1:
      # 获取tag类型
      start_last_tag = result[start_index:start_index+2]
      if(start_last_tag == "{{"):
        # 变量替换
        end_index = result.find('}}', start_index)
        key = result[start_index + 2:end_index].strip()
        #根据key获取value
        value = self.get_value(key,data)
        #拼接
        result = result[:start_index] + str(value) + result[end_index + 2:]
      # 对if或者for进行处理
      elif(start_last_tag == "{%"):
        end_index = result.find('%}', start_index)
        tagbody = result[start_index+2:end_index]
        tagFragment = tagbody.split(" ")
        while '' in tagFragment:
          tagFragment.remove('')
        type = tagFragment[0]
        # 处理if
        if(type==self.IF):
          # endTag = "{% endif %}"
          # try:
          #   endTag=re.compile(r"{%\s*endif\s*%}").search(result).group(0)
          # except:
          #   raise Exception('Missing end tag in '+ tagbody)
          # endTagIndex=result.find(endTag)
          endTagIndex,endTag = self.findEndTag(result[start_index:],self.IF)
          endTagIndex = endTagIndex+start_index
          #得到条件
          tagFragment[0]=""
          condition = "".join(tagFragment)
          print(condition)
          exeConditionRes=self.conditionalExecution(data,condition)
          print(exeConditionRes)
          #判断结果
          if(exeConditionRes):
            # 先删除end标签保证start标签起始坐标不变
            result = result[:endTagIndex] + result[endTagIndex + len(endTag):]
            result = result[:start_index] + result[end_index + 2:]
          else:
            result = result[:start_index] + result[endTagIndex + len(endTag):]
        # 处理for
        elif(type==self.FOR):
          # 定位end标签的位置
          # endTag = "{% endfor %}"
          # try:
          #   endTag=re.compile(r"{%\s*endfor\s*%}").search(result).group(0)
          # except:
          #   raise Exception('Missing end tag in '+ tagbody)
          # endTagIndex=result.find(endTag)
          endTagIndex,endTag = self.findEndTag(result[start_index:],self.FOR)
          endTagIndex = endTagIndex+start_index
          tagFragment[0]=""
          # 获取被循环的字典值
          value = self.get_value(tagFragment[-1],data)
          print(value)
          resBody = ""
          for item in value:
            t = TemplateEngine()
            forBody = result[end_index+2:endTagIndex]
            t.load_template(forBody)
            data[tagFragment[1]] = item
            res = t.render_template(data)
            resBody = resBody + res
            data.pop(tagFragment[1])
          result = result[:start_index] + resBody + result[endTagIndex + len(endTag):]

      start_index = self.findStartTag(result)
      print("====")
    return result

# 找到开始标签的位置
  def findStartTag(self,msg:str):
    valueTagIndex = msg.find('{{')
    logicTagIndex = msg.find('{%')
    if(valueTagIndex == -1 and logicTagIndex == -1):
      return -1
    elif(valueTagIndex == -1):
      return logicTagIndex
    elif(logicTagIndex == -1):
      return valueTagIndex
    else:
      return  valueTagIndex < logicTagIndex and valueTagIndex or logicTagIndex

# 找到结束标签的位置
  def findEndTag(self,body:str,type):
    level = 1
    position = -1
    tag = ""
    startTagRe = ""
    endTagRe = ""
    if(type == self.IF):
      startTagRe = "{%\s*if.*%}"
      endTagRe = "{%\s*endif\s*%}"
    elif(type == self.FOR):
      startTagRe = "{%\s*for.*%}"
      endTagRe = "{%\s*endfor\s*%}"
    # 初始化开始标签
    startTag = re.compile(startTagRe).search(body).group(0)
    startTagIndex = body.find(startTag)
    body = body[startTagIndex+len(startTag):]
    position = startTagIndex+len(startTag)
    while(level>0):
      # 判断最近的Tag是开始标签还是结束标签
      try:
        startTag = re.compile(startTagRe).search(body).group(0)
        startTagIndex = body.find(startTag)
      except:
        startTagIndex = -1
      try:
        endTag = re.compile(endTagRe).search(body).group(0)
        endTagIndex = body.find(endTag)
      except:
        endTagIndex = -1
      if(endTagIndex ==-1):
        # 错误，没有结束标签
        return -1,-1
      # 如果开始标签不存在，结束标签存在或结束标签在开始标签之前  leve-1 并记录结束标签的位置和tag 
      elif(startTagIndex==-1 or endTagIndex < startTagIndex):
        level = level-1
        tag = endTag
        catPosition = endTagIndex+len(endTag)
        position = position + catPosition
        body = body[catPosition:]
        if(level==0):
          position = position - len(endTag)
          break
      elif(startTagIndex<endTagIndex):
        level = level+1
        catPosition = startTagIndex+len(startTag)
        position = position + catPosition
        body = body[startTagIndex+len(startTag):]
    
    return position,tag
      
        
      

  # 判断条件表达式
  def conditionalExecution(self,data: dict,condition:str):
    #替换变量
    ## 截取字符串
    conditionList = re.split(self.delimiters, condition)
    ## 替换变量值
    for i in range(len(conditionList)):
      if(conditionList[i].find("'")==-1 and conditionList[i].find('"')):
        try:
          value=self.get_str_value(conditionList[i],data)
          conditionList[i] = value
        except:
          pass
    exeCondition=" ".join(conditionList)
    print(exeCondition)
    return eval(exeCondition)


# 获取字符串结果
  def get_str_value(self,key: str,data: dict):
    value = self.getRealValue(key,data)
    if isinstance (value,str):
      value = "'"+value+"'" 
    elif(not isinstance(value,dict)):
      value = str(value)
    
    return value 

# 获取字典中的值
  def get_value(self,key: str,data: dict):
    value = self.getRealValue(key,data)
    if isinstance (value,bool):
        value = str(value).lower()
    return value

# 获取真实值
  def getRealValue(self,key: str,data: dict):
    keys = key.split('.')
    value = data
    for k in keys:
      start = k.find("[")
      # 如果是数组
      if(start!=-1):
        end = k.find("]")
        try:
          sp =k[start+1:end]
        except:
          raise Exception('ERROR Array indices have no end flags')
        try:
          index=int(sp)
        except:
          raise Exception('ERROR Array indices are not numbers')
        try:
          value = value[k[:start]][index]
        except:
          raise Exception('ERROR Array index out of bounds')
      # 不是数组
      else:
        try:
          value = value[k]
        except:
          raise Exception('ERROR Variable does not exist')
    return value
      
      






# 服务bean
class ServiceBean:
  # 服务实现的包名，不同版本保持一致
  url = None

  # 名称
  name = None

  # 版本
  versionCode = None

  # 能力 K:能力id   V: CapabilityBean
  capabilityBeanDict = {}
  
  # 事件 k:事件id  V: EventBean
  eventBeanDict = {}

  # 文件所在路径
  filePath = None
  
  def __init__(self) -> None:
    self.capabilityBeanDict = {}

  def __str__(self):
    capabilityStr = "{"
    for key, value in self.capabilityBeanDict.items():
      capabilityStr += "\n \t\t "+value.__str__()
    eventBeanStr = "{"
    for key, value in self.eventBeanDict.items():
      eventBeanStr += "\n \t\t "+value.__str__()
    return f"\nServiceBean(filePath={self.filePath},url={self.url}, name={self.name}, versionCode={self.versionCode}, \n \t capabilityBeanDict={capabilityStr} ,\n \t eventBeanDict = {eventBeanStr}) \n"
 


from ..core.scheduleService import ScheduleService
from ..systemService.mqttService import MqttService
from .loader.hardwareInfoLoader import HardwareInfoLoader
from ..import config
from datetime import datetime


# 代理加载存储字典
proxyDict:dict = None

# 服务加载存储字典
serviceDict:dict = None

stateInfo:dict = None

# mqtt客户端
pervasiveMqttClient:MqttService = None


# 定时器
taskServer:ScheduleService = None

# 硬件信息加载器
hardwareInfoLoader:HardwareInfoLoader = None

"""
* 返回service信息列表，每个节点是一个字典格式：
* {
*    'versionCode':服务版本号，
*    'id':服务id,
*    'serviceId':服务id,与上面id等价
* }
"""
def getServiceListInfo():
    services = []
    for key in serviceDict.keys():
        serviceItem= serviceDict[key].__dict__
        serviceItem =  {
            'versionCode': serviceItem['versionCode'],
            'id': serviceItem['url'],
            'serviceId': serviceItem['url']
        }
        services.append(serviceItem)
    return services

"""
* 返回service列表，每个节点是一个ServiceBean
"""
def getServiceList():
    services = []
    for key in serviceDict.keys():
        serviceItem = serviceDict[key]
        services.append(serviceItem)
    return services


def getSystemInfo():
    current_time = datetime.now()
    formatted_time = current_time.strftime("%Y-%m-%dT%H:%M:%S")
    resDict = {
        "deviceId":config.DEVICE_ID,
        "dateTime":formatted_time,
        "stateInfo":stateInfo,
        "hardwareInfo":hardwareInfoLoader.getHardwareInfo()
    }
    return resDict



import setuptools #导入setuptools打包工具
 
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
 
setuptools.setup(
    name="iotPervasiveServiceSDK", # 用自己的名替换其中的YOUR_USERNAME_
    version="0.0.19",    #包版本号，便于维护版本
    author="whu贾向阳团队-葛家和",    #作者，可以写自己的姓名
    author_email="2898534520@qq.com",    #作者联系方式，可写自己的邮箱地址
    description="设备端直连框架python",#包的简述
    long_description=long_description,    #包的详细介绍，一般在README.md文件内
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),  # Make sure to include 'where' argument
    zip_safe=False,  # Disable zip-safe mode for easier debugging
    include_package_data=True, 
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",  
    ],
    python_requires='>=3.6',    #对python的最低版本要求
    install_requires=[  #对应依赖信息
        "bottle==0.12.25",
        "netifaces==0.11.0",
        "paho_mqtt==1.6.1",
        "psutil==5.9.4",
        "Requests>=2.0.0",
        "ujson"
    ],
)


import requests
import os
from .. import config
from ..core.loader.serversLoader import ServiceLoader
from ..core import applicationContext
from ..pojo import serviceBean
from ..utils import fileUtils
import sys
from ..utils.tarGzUtils import TarGzUtils
from ..utils import logutil
from ..core import applicationContext
from ..core.exception.serviceException import ServiceNotExistException
from ..utils import pathUtils as PathUtil

SERVICE_TARGZ_FILE_NAME = "serviceImpl.tar.gz"

log = logutil.get_logger(__name__)
'''
泛在服务平台服务管理
'''

'''
删除指定服务
'''
def deleteService(id:str)->bool:
    
    # 获取服务信息    
    service:serviceBean = applicationContext.serviceDict.get(id,None)
    # 服务不存在则 
    if (service == None):
        return True
    # 服务路径拼接    
    servicePath =  PathUtil.joinPath(config.BASE_PATH,get_service_path(config.SERVICE_PATH,service.filePath))
    # 删除服务    
    fileUtils.removePath(servicePath)
    # 重新加载服务列表    
    applicationContext.serviceDict=ServiceLoader.LoadServiceDict(config.SERVICE_RELATIVE_PATH,config.BASE_PATH)
    # 检查是否删除
    service:serviceBean = applicationContext.serviceDict.get(id,None)
    if (service == None):
        return True
    else:
        return False





'''
* 增加/更新服务
'''
def updateService(id:str):
    downloadService(id)
    applicationContext.serviceDict=ServiceLoader.LoadServiceDict(config.SERVICE_RELATIVE_PATH,config.BASE_PATH)
    



'''
* 服务下载
'''
def downloadService(id:str):
    # 获取服务文件信息
    try :
        getUrl = config.BASE_URL+"/serviceImpl/"+id+"/getFolderPath"
        log.info(getUrl)
        res = requests.get(getUrl)
    except Exception as e:
        raise Exception('rest get serviceImplInfo error ')
    if(res.status_code!=200):
        raise Exception('rest get error,status_code = '+ str(res.status_code))
    serviceFilePath = res.json()["data"]
    log.info(serviceFilePath)
    # 拼接服务路径   判断服务是否存在 存在则采用原路径
    existingService = applicationContext.serviceDict.get(id.split("::")[0],None)
    if existingService!=None:
        deleteService(id.split("::")[0])
    servicePath = PathUtil.joinPath(config.SERVICE_PATH,"service"+''.join([('0'+hex(ord(os.urandom(1)))[2:])[-2:] for x in range(8)]))
    log.info("servicePath"+servicePath)
    serviceTargzPath =  PathUtil.joinPath(servicePath,SERVICE_TARGZ_FILE_NAME)
    # 下载压缩包    
    downloadServiceFile(serviceFilePath,serviceTargzPath)
    # 解压文件
    tarFile = TarGzUtils(serviceTargzPath) 
    try:
        tarFile.extractAndDeleteToDir(servicePath)
    except Exception as e:
        log.error(str(e))
    
    
'''
 下载服务文件
'''
def downloadServiceFile(filePath:str,servicePath:str):
    downloadUrl=config.BASE_URL+"/file/folder?urlPath="+filePath+"&type=tar.gz"
    log.info("url:"+downloadUrl)
    log.info("service:"+servicePath)
    try:
        fileUtils.downloadFile(downloadUrl,servicePath)
    except Exception as e:
        raise Exception('下载时出现异常: url:'+downloadUrl+",filePath:"+servicePath+ "err:" +str(e))


def get_service_path(base_path, config_path):
    base_dir = os.path.basename(base_path)
    rel_path = os.path.relpath(config_path, base_path)
    service_path = os.path.join(base_dir, rel_path.split(os.sep)[0])
    return service_path





from .utils import pathUtils as PathUtil

BASE_NAME = "SDK"

# 设备实例id
DEVICE_ID = "UnknownDevice"

# topic 
INCOCATION_ENUM = 'invocation' 
CALLBACK_ENUM = 'callback'
UPDATE_ENUM = 'update'
# 泛在服务平台TOPIC
##接收
# 服务调用
INVOCATION_SERVICE_EXECUTE_TOPIC = ''
# 服务调用回调消息
INVOCATION_SERVICE_CALLBACK_TOPIC = ''
# 服务更新回调消息
UPDATE_SERVICE_CALLBACK_TOPIC = ''
# 更新服务topic
UPDATE_SERVICE_TOPIC = ''

## 上报
# 事件触发topic
ON_EVENT_TOPIC = ''
# 情境上报topic
UPLOAD_CONTEXT_TOPIC = ''
# 服务列表上报topic
UPLOAD_SERVICE_LIST_TOPIC = ''
# 设备硬件信息上报
UPLOAD_HARDWARE_TOPIC = ''

# mqtt服务器相关配置
MQTT_ADDRESS="116.204.102.248"
MQTT_PORT=1883

# 基础路径
BASE_PATH="/"
SERVICE_PATH = "/service"
SERVICE_RELATIVE_PATH = "service"
PROXY_PATH = "/proxy"


# url相关
BASE_URL = "http://192.168.0.103:8084"

# 初试化TOPIC
def initTopic(deviceId:str):
  global DEVICE_ID
  DEVICE_ID = deviceId
  global INVOCATION_SERVICE_EXECUTE_TOPIC
  INVOCATION_SERVICE_EXECUTE_TOPIC='/client/invocation/%s/+/+'%(DEVICE_ID)
  global INVOCATION_SERVICE_CALLBACK_TOPIC
  INVOCATION_SERVICE_CALLBACK_TOPIC = '/server/invocation/callback/%s/${service}/${capability}/${msgid}'%(DEVICE_ID)
  global UPDATE_SERVICE_CALLBACK_TOPIC
  UPDATE_SERVICE_CALLBACK_TOPIC = '/server/update/callback/${service}/${capability}/${msgid}'
  global UPDATE_SERVICE_TOPIC
  UPDATE_SERVICE_TOPIC='/client/%s/%s/+'%(UPDATE_ENUM,DEVICE_ID)

  global ON_EVENT_TOPIC
  ON_EVENT_TOPIC='/server/event/%s'%(DEVICE_ID)
  global UPLOAD_CONTEXT_TOPIC
  UPLOAD_CONTEXT_TOPIC='/server/context/%s'%(DEVICE_ID)
  global UPLOAD_SERVICE_LIST_TOPIC
  UPLOAD_SERVICE_LIST_TOPIC = '/server/deviceInfo/%s'%(DEVICE_ID)
  global UPLOAD_HARDWARE_TOPIC
  UPLOAD_HARDWARE_TOPIC = '/server/deviceHardwareInfo/%s'%(DEVICE_ID)
  

# 初始化路径信息 
def initPath(basePath:str,servicePath:str,proxyPath:str):
    global BASE_PATH
    BASE_PATH = basePath
    global SERVICE_RELATIVE_PATH
    SERVICE_RELATIVE_PATH = servicePath
    global SERVICE_PATH
    SERVICE_PATH =  PathUtil.joinPath(basePath,servicePath)
    global PROXY_PATH
    PROXY_PATH = PathUtil.joinPath(basePath,proxyPath)
    



from bottle import Bottle, run, request, template
from ..core.serverInvoker import CapabilityInvoker
import json
from ..utils import TimeUtil
import threading

app = Bottle()

@app.route('/client/invocation/<deviceId>/<serviceId>/<capabilityId>', method='POST')
# web局域网服务调用
def _httpHandlerInvokerPost(deviceId, serviceId, capabilityId):
    print("======rest server invoke======")
    msgJson = request.json.get("param","")
    runSuccess = False
    res = None
    try:
        res = CapabilityInvoker.invokeCapability(
            serviceId, capabilityId, msgJson
        )
        runSuccess = True
    except Exception as e:
        runSuccess = False
        res = e.__str__()
    # 拼接msg
    msgDict = {
        "success": runSuccess,
        "data": res,
        "timestamp": TimeUtil.getTimerMs(),
    }
    print(msgDict)
    return template(json.dumps(msgDict, ensure_ascii=False), name=__name__);

def _runServer():
    run(app, host='localhost', port=1024, debug=True)

def runWebServer():
        # 创建一个新线程来运行 Bottle 服务
    server_thread = threading.Thread(target=_runServer)
    server_thread.start()
    

